-- name: create_tables#
-- Initialize the core db tables. Arguably migration 0.
----------------------------------------------------------------------------------------------------
-- User structures
CREATE TABLE IF NOT EXISTS groups (
   id INTEGER PRIMARY KEY AUTOINCREMENT
 , name TEXT
 , priority INTEGER CHECK(priority IS NOT NULL AND priority > 0)
);

INSERT OR IGNORE INTO groups (id, name, priority) VALUES (0, 'root', 20);
INSERT OR IGNORE INTO groups (id, name, priority) VALUES (1, 'users', 10);
INSERT OR IGNORE INTO groups (id, name, priority) VALUES (2, 'guests', 0);

CREATE TABLE IF NOT EXISTS user_statuses (
   id INTEGER PRIMARY KEY AUTOINCREMENT
 , name TEXT
 , UNIQUE(name)
);

INSERT OR IGNORE INTO user_statuses (id, name) VALUES (-3, 'unverified');
INSERT OR IGNORE INTO user_statuses (id, name) VALUES (-2, 'unapproved');
INSERT OR IGNORE INTO user_statuses (id, name) VALUES (-1, 'disabled');
INSERT OR IGNORE INTO user_statuses (id, name) VALUES (1, 'enabled');

CREATE TABLE IF NOT EXISTS users (
   id INTEGER PRIMARY KEY AUTOINCREMENT
 , group_id INTEGER
 , name TEXT
 , email TEXT
 , hash TEXT
 , status_id INTEGER
 , verification_token TEXT DEFAULT (lower(hex(randomblob(32))))
 , verified_at TEXT
 , approved_at TEXT
 , enabled_at TEXT
 , FOREIGN KEY(group_id) REFERENCES groups(id)
 , FOREIGN KEY(status_id) REFERENCES user_statuses(id)
 , UNIQUE(name)
 , UNIQUE(email)
);

----------------------------------------------------------------------------------------------------
-- Keys represent API keys and auth sessions
CREATE TABLE IF NOT EXISTS user_keys (
   id TEXT PRIMARY KEY DEFAULT (lower(hex(randomblob(32))))
 , user_id INTEGER
 , name TEXT
 , expiration TEXT
);

----------------------------------------------------------------------------------------------------
-- Printers represent physical devices (really octoprint API targets) that jobs can go to
CREATE TABLE IF NOT EXISTS printer_statuses (
   id INTEGER PRIMARY KEY AUTOINCREMENT
 , name TEXT
 , UNIQUE(name)
);

INSERT OR IGNORE INTO printer_statuses (id, name) VALUES (-1, 'error');
INSERT OR IGNORE INTO printer_statuses (id, name) VALUES (0, 'disconnected');
INSERT OR IGNORE INTO printer_statuses (id, name) VALUES (2, 'connecting');
INSERT OR IGNORE INTO printer_statuses (id, name) VALUES (2, 'connected');
INSERT OR IGNORE INTO printer_statuses (id, name) VALUES (3, 'idle');
INSERT OR IGNORE INTO printer_statuses (id, name) VALUES (4, 'running');

CREATE TABLE IF NOT EXISTS printers (
   id INTEGER PRIMARY KEY AUTOINCREMENT
 , name TEXT
 , url TEXT
 , stream_url TEXT
 , api_key TEXT
 , status_id INTEGER
 , last_poll_date TEXT
 , FOREIGN KEY(status_id) REFERENCES printer_statuses(id)
);

----------------------------------------------------------------------------------------------------
-- Files are printables
CREATE TABLE IF NOT EXISTS files (
   id INTEGER PRIMARY KEY AUTOINCREMENT
 , user_id INTEGER
 , filename TEXT
 , path TEXT
 , upload_date TEXT
 , FOREIGN KEY(user_id) REFERENCES user(id)
);

----------------------------------------------------------------------------------------------------
-- A job is a request for a copy of a file to be run off
-- For simplicity, jobs also serve as scheduling records
CREATE TABLE IF NOT EXISTS jobs (
   id INTEGER PRIMARY KEY AUTOINCREMENT
 , user_id INTEGER NOT NULL
 , file_id INTEGER NOT NULL
 , priority INTEGER CHECK(priority IS NOT NULL AND 0 <= priority)
 , started_at TEXT
 , cancelled_at TEXT
 , finished_at TEXT
 , state TEXT
 , message TEXT
 , printer_id INTEGER
 , FOREIGN KEY(user_id) REFERENCES users(id)
 , FOREIGN KEY(file_id) REFERENCES files(id)
 , FOREIGN KEY(printer_id) REFERENCES printer(id)
);

----------------------------------------------------------------------------------------------------
-- Emails are used for notifications
CREATE TABLE IF NOT EXISTS email_spool (
   id INTEGER PRIMARY KEY AUTOINCREMENT
 , user_id INTEGER NOT NULL
 , subject TEXT
 , body TEXT
 , sent_at TEXT
 , FOREIGN KEY(user_id) REFERENCES users(id)
);

----------------------------------------------------------------------------------------------------
-- Users
----------------------------------------------------------------------------------------------------

-- name: try-create-user^
INSERT INTO users (
   name
 , email
 , hash
 , group_id
 , status_id
)
VALUES (:name, :email, :hash, :gid, :sid)
RETURNING *
;

-- name: fetch_user^
SELECT
   *
FROM users
WHERE
   id = :uid
;

-- name: list-users
SELECT
   id
 , email
FROM users
;

-- name: list-unverified-users
SELECT
   *
FROM users
WHERE
   status_id = -2
   AND verified_at IS NULL
;

-- name: verify-user!
UPDATE users
SET
   verified_at = datetime('now')
WHERE
   id = :uid
;


-- name: set-user-status!
UPDATE users
SET
   status_id = :sid
WHERE
   id = :uid
;

----------------------------------------------------------------------------------------------------
-- User statuses
----------------------------------------------------------------------------------------------------

-- name: fetch-user-status^
SELECT
   id
 , name
FROM user_statuses
WHERE
   id = :uid
;

----------------------------------------------------------------------------------------------------
-- Sessions / 'keys'
----------------------------------------------------------------------------------------------------

-- name: create-key^
INSERT INTO user_keys (
   user_id
 , name
 , expiration
)
VALUES (:uid, :name, :expiration)
RETURNING
  id
;

-- name: try-login^
SELECT
   id
 , status_id
FROM users
WHERE
   (name = :username AND hash = :hash)
   OR (email = :username AND hash = :hash)
LIMIT 1
;

-- name: list-keys
SELECT
   id
 , name
 , expiration
FROM user_keys
WHERE
   user_id = :uid
;

-- name: fetch-key^
SELECT
   *
FROM user_keys
WHERE
   id = :kid
;

-- name: try-key^
SELECT
   user_id
FROM user_keys
WHERE
   (expiration IS NULL OR unixepoch(expiration) > unixepoch('now'))
   AND id = :kid
;

-- name: refresh-key
UPDATE user_keys
SET
   expiration = :expiration
WHERE
   id = :kid
;

-- name: delete-key
DELETE FROM user_keys
WHERE
   user_id = :uid
   AND id = :kid
;

----------------------------------------------------------------------------------------------------
-- Printers
----------------------------------------------------------------------------------------------------

-- name: try-create-printer^
INSERT INTO printers (
   name
 , url
 , api_key
 , status_id
)
VALUES (:name, :url, :api_key, :status_id)
RETURNING
   id
;

-- name: fetch-printer^
SELECT
   p.id
 , p.name
 , p.url
 , p.api_key
 , p.last_poll_date
 , s.name as status
FROM printers p
INNER JOIN printer_statuses s ON p.status_id = s.id
WHERE p.id = :pid
;

-- name: list-printers
SELECT
   p.id
 , p.name
 , p.url
 , p.api_key
 , p.last_poll_date
 , s.name as status
FROM printers p
INNER JOIN printer_statuses s ON p.status_id = s.id
;

-- name: list-idle-printers
SELECT p.id
FROM printers p
LEFT JOIN (SELECT id, printer_id FROM jobs WHERE finished_at IS NULL) j
   ON p.id = j.printer_id
INNER JOIN printer_statuses s
   ON p.status_id = s.id
WHERE
   j.id IS NULL
   AND s.name = 'idle'
;

-- name: update-printer-status!
UPDATE printers
SET
   status_id = (SELECT id FROM printer_statuses WHERE name = :status)
 , last_poll_date = datetime('now')
WHERE
   id = :uid
;

----------------------------------------------------------------------------------------------------
-- Files
----------------------------------------------------------------------------------------------------

-- name: create-file^
INSERT INTO files (
   user_id
 , filename
 , path
 , upload_date
)
VALUES (:uid, :filename, :path , datetime('now'))
RETURNING
   id
;

-- name: list-files
SELECT
   *
FROM files
WHERE
   user_id = :uid
;

-- name: delete-file!
DELETE FROM files
WHERE
   user_id = :uid
   AND id = :fid
;

-- name: fetch-file^
SELECT
   *
FROM files
WHERE
   user_id = :uid
   AND id = :fid
;

----------------------------------------------------------------------------------------------------
-- Jobs
----------------------------------------------------------------------------------------------------

-- name: create-job^
INSERT INTO jobs (
   user_id
 , file_id
 , priority
)
VALUES (
   :uid
 , :fid,
 , (
     SELECT priority + :priority
     FROM users
     WHERE uid = :uid
    )
)
RETURNING
   id
;
-- name: fetch-job^
SELECT
   *
FROM jobs
WHERE
   user_id = :uid
   AND id = :fid
;

-- name: list-jobs
SELECT
   *
FROM jobs
WHERE
   (:uid IS NULL OR user_id = :uid)
;

-- name: list-jobs-by-file
SELECT
   *
FROM jobs
WHERE
   file_id = :fid
;

-- name: list-job-queue
SELECT
   *
FROM jobs
WHERE
   finished_at IS NULL
   AND (:uid IS NULL OR user_id = :uid)
ORDER BY
   priority DESC
;

-- name: poll-job-queue
SELECT
   *
FROM jobs
WHERE
   started_at IS NULL
   AND finished_at IS NULL
   AND printer_id IS NULL
ORDER BY
   priority DESC
LIMIT 1
;

-- name: list-job-history
SELECT
   *
FROM jobs
WHERE
   finished_at IS NOT NULL
   AND (:uid IS NULL OR user_id = :uid)
ORDER BY
   datetime(finished_at) DESC
;

-- name: list-mapped-jobs
SELECT
   *
FROM jobs
WHERE
   started_at IS NULL
   AND printer_id IS NOT NULL
;

-- name: list-running-jobs
SELECT
   *
FROM jobs
WHERE
   started_at IS NOT NULL
   AND printer_id IS NOT NULL
   AND finished_at IS NULL
;

-- name: list-canceling-jobs
SELECT
   *
FROM jobs
WHERE
   finished_at IS NULL
   AND cancelled_at IS NOT NULL
;

-- name: fetch-job-by-printer^
SELECT
   *
FROM jobs
WHERE
   printer_id = :pid
   AND finished_at IS NULL
;

-- name: assign-job!
UPDATE jobs
SET
   printer_id = :pid
WHERE
   id = :jid
;

-- name: start-job!
UPDATE jobs
SET
   started_at = datetime('now')
WHERE
   id = :jid
;

-- name: cancel-job!
UPDATE jobs
SET
   cancelled_at = datetime('now')
WHERE
   user_id = :uid
   AND id = :jid
;

-- name: finish-job!
UPDATE jobs
SET
   finished_at = datetime('now')
 , state = :state
 , message = :message
WHERE
   id = :jid
;

-- name: delete-job!
DELETE FROM jobs
WHERE
   user_id = :uid
   AND id = :jid
;

-- name: create-email!
INSERT INTO email_spool (
   user_id
 , subject
 , body
) VALUES (
   :uid
 , :subject
 , :body
)
RETURNING
   id
;

-- name: send-email!
UPDATE email_spool
SET
   sent_at = datetime('now')
WHERE
   id = :eid
;

-- name: poll-email-queue
SELECT
   s.id as `id`
 , u.email as `to`
 , subject
 , body
FROM email_spool s
INNER JOIN users u
   ON s.user_id = u.id
WHERE
   s.sent_at IS NULL
LIMIT 1
;
